# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import re
from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.camera import (
    CameraAction,
    CameraBarcodeEvent,
    CameraEvent,
    CameraReportBarcodeAction,
    CameraStartViewfinderAction,
    CameraStartViewfinderEvent,
    CameraState,
    CameraStopViewfinderEvent,
    InputDescription,
)
from ubo_app.store.services.keypad import Key, KeypadAction

Action = InitAction | CameraAction


def pop_queue(state: CameraState) -> CameraState:
    if len(state.queue) > 0:
        input_description, *queue = state.queue
        return replace(state, current=input_description, queue=queue)
    return replace(
        state,
        is_viewfinder_active=False,
        current=None,
    )


def reducer(  # noqa: C901
    state: CameraState | None,
    action: Action,
) -> ReducerResult[CameraState, Action, CameraEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return CameraState(is_viewfinder_active=False, queue=[])
        raise InitializationActionError(action)

    if isinstance(action, CameraStartViewfinderAction):
        if state.is_viewfinder_active:
            return replace(
                state,
                queue=[
                    *state.queue,
                    InputDescription(id=action.id, pattern=action.pattern),
                ],
            )
        return CompleteReducerResult(
            state=replace(
                state,
                is_viewfinder_active=True,
                current=InputDescription(id=action.id, pattern=action.pattern),
            ),
            events=[CameraStartViewfinderEvent(pattern=action.pattern)],
        )

    if isinstance(action, CameraReportBarcodeAction) and state.current:
        for code in action.codes:
            if state.current.pattern:
                match = re.match(state.current.pattern, code)
                if match:
                    return CompleteReducerResult(
                        state=pop_queue(state),
                        events=[
                            CameraBarcodeEvent(
                                id=state.current.id,
                                code=code,
                                group_dict=match.groupdict(),
                            ),
                            CameraStopViewfinderEvent(id=None),
                        ],
                    )
            else:
                return CompleteReducerResult(
                    state=pop_queue(state),
                    events=[
                        CameraBarcodeEvent(
                            id=state.current.id,
                            code=code,
                            group_dict=None,
                        ),
                        CameraStopViewfinderEvent(id=None),
                    ],
                )

            return state

    if isinstance(action, KeypadAction):  # noqa: SIM102
        if action.key == Key.BACK:
            return CompleteReducerResult(
                state=state,
                events=[
                    CameraStopViewfinderEvent(
                        id=state.current.id if state.current else None,
                    ),
                ],
            )

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
import math

WIFI_STATE_ICON_ID = 'wifi:state'
WIFI_STATE_ICON_PRIORITY = -12
SIGNAL_ICONS = ['󰤯', '󰤟', '󰤢', '󰤥', '󰤨']


def get_signal_icon(strength: float) -> str:
    return SIGNAL_ICONS[math.floor(strength / 100 * 4.999)]



"""Signature of the `register_service` function."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Coroutine, Protocol, TypeAlias

if TYPE_CHECKING:
    from redux import ReducerType


class Service(Protocol):
    """Utilities to initialize a service."""

    def register_reducer(self: Service, reducer: ReducerType) -> None:
        """Register a reducer."""


SetupFunction: TypeAlias = (
    Callable[[Service], Coroutine | None] | Callable[[], Coroutine | None]
)


def register(
    *,
    service_id: str,
    label: str,
    setup: SetupFunction,
) -> None:
    """Register a service, meant to be called in `ubo_handle.py` file of services."""
    _ = service_id, label, setup
    msg = 'This function is not meant to be called out of the services'
    raise NotImplementedError(msg)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction, BaseEvent


class CameraAction(BaseAction): ...


class CameraStartViewfinderAction(CameraAction):
    id: str
    pattern: str | None


class CameraEvent(BaseEvent): ...


class CameraStartViewfinderEvent(CameraEvent):
    pattern: str | None


class CameraStopViewfinderEvent(CameraEvent):
    id: str | None


class CameraReportBarcodeAction(CameraAction):
    codes: list[str]


class CameraBarcodeEvent(CameraEvent):
    id: str | None
    code: str
    group_dict: dict[str, str | None] | None


class InputDescription(Immutable):
    id: str
    pattern: str | None


class CameraState(Immutable):
    current: InputDescription | None = None
    is_viewfinder_active: bool
    queue: list[InputDescription]



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction


class LightDMAction(BaseAction): ...


class LightDMUpdateStateAction(LightDMAction):
    is_active: bool | None = None
    is_enabled: bool | None = None


class LightDMClearEnabledStateAction(LightDMAction): ...


class LightDMState(Immutable):
    is_active: bool = False
    is_enabled: bool | None = None



# pyright: reportMissingImports=false
"""Provides the system bus for the application."""

from __future__ import annotations

from threading import current_thread

from sdbus import SdBus, sd_bus_open_system, set_default_bus

system_buses = {}


def get_system_bus() -> SdBus:
    """Get the system bus for current thread."""
    thread = current_thread()
    if thread not in system_buses:
        system_buses[thread] = sd_bus_open_system()
    set_default_bus(system_buses[thread])
    return system_buses[thread]



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_app.services import Service, register


def setup(service: Service) -> None:
    from reducer import reducer
    from setup import init_service

    service.register_reducer(reducer)

    init_service()


register(
    service_id='lightdm',
    label='LightDM',
    setup=setup,
)



# pyright: reportMissingImports=false
# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

import asyncio
import contextlib
import uuid
from typing import TYPE_CHECKING, Any, Coroutine, TypeVar, cast

from debouncer import DebounceOptions, debounce
from ubo_gui.constants import DANGER_COLOR

from ubo_app.store import dispatch
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.wifi import (
    ConnectionState,
    GlobalWiFiState,
    WiFiConnection,
    WiFiType,
)
from ubo_app.utils.bus_provider import get_system_bus

if TYPE_CHECKING:
    from asyncio.tasks import _FutureLike

RETRIES = 3

T = TypeVar('T')


def wait_for(task: _FutureLike[T]) -> Coroutine[Any, Any, T]:
    return asyncio.wait_for(task, timeout=10.0)


from sdbus_async.networkmanager import (  # noqa: E402
    AccessPoint,
    ActiveConnection,
    DeviceState,
    NetworkConnectionSettings,
    NetworkDeviceGeneric,
    NetworkDeviceWireless,
    NetworkManager,
    NetworkManagerConnectionProperties,
    NetworkManagerSettings,
)
from sdbus_async.networkmanager.enums import (  # noqa: E402
    ConnectionState as SdBusConnectionState,
)
from sdbus_async.networkmanager.enums import DeviceType  # noqa: E402


async def get_wifi_device() -> NetworkDeviceWireless | None:
    network_manager = NetworkManager(get_system_bus())
    devices_paths = await wait_for(
        network_manager.get_devices(),
    )
    for device_path in devices_paths:
        generic_device = NetworkDeviceGeneric(device_path, get_system_bus())
        if (
            await wait_for(
                generic_device.device_type,
            )
            == DeviceType.WIFI
        ):
            return NetworkDeviceWireless(device_path, get_system_bus())
    return None


async def get_wifi_device_state() -> GlobalWiFiState:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return GlobalWiFiState.UNKNOWN

    state = await wifi_device.state
    if state is DeviceState.UNKNOWN:
        return GlobalWiFiState.UNKNOWN
    if state in (
        DeviceState.DISCONNECTED,
        DeviceState.UNMANAGED,
        DeviceState.UNAVAILABLE,
        DeviceState.FAILED,
    ):
        return GlobalWiFiState.DISCONNECTED
    if state in (DeviceState.NEED_AUTH,):
        return GlobalWiFiState.NEEDS_ATTENTION
    if state in (
        DeviceState.DEACTIVATING,
        DeviceState.PREPARE,
        DeviceState.CONFIG,
        DeviceState.IP_CONFIG,
        DeviceState.IP_CHECK,
        DeviceState.SECONDARIES,
    ):
        return GlobalWiFiState.PENDING
    if state == DeviceState.ACTIVATED:
        return GlobalWiFiState.CONNECTED

    return GlobalWiFiState.UNKNOWN


@debounce(wait=0.5, options=DebounceOptions(trailing=True, time_window=2))
async def request_scan() -> None:
    wifi_device = await get_wifi_device()
    if wifi_device:
        await wait_for(wifi_device.request_scan({}))


async def get_access_points() -> list[AccessPoint]:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return []

    access_points = await wait_for(
        wifi_device.access_points,
    )
    return [
        AccessPoint(access_point_path, get_system_bus())
        for access_point_path in access_points
    ]


async def get_active_access_point() -> AccessPoint | None:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return None

    active_access_point = await wait_for(
        wifi_device.active_access_point,
    )
    if not active_access_point or active_access_point == '/':
        return None

    return AccessPoint(active_access_point, get_system_bus())


async def get_active_access_point_ssid() -> str | None:
    active_access_point = await get_active_access_point()
    if not active_access_point:
        return None

    return (
        await wait_for(
            active_access_point.ssid,
        )
    ).decode('utf-8')


async def get_active_connection() -> ActiveConnection | None:
    wifi_device = await get_wifi_device()
    if wifi_device is None:
        return None

    active_connection = await wait_for(
        wifi_device.active_connection,
    )
    if not active_connection or active_connection == '/':
        return None

    return ActiveConnection(active_connection, get_system_bus())


async def get_active_connection_ssid() -> str | None:
    active_connection = await get_active_connection()
    if not active_connection:
        return None

    connection = NetworkConnectionSettings(await active_connection.connection)
    settings = await connection.get_settings()
    return settings['802-11-wireless']['ssid'][1].decode('utf-8')


async def get_saved_ssids() -> list[str]:
    network_manager_settings = NetworkManagerSettings(get_system_bus())
    connections = [
        NetworkConnectionSettings(i, get_system_bus())
        for i in await wait_for(
            network_manager_settings.connections,
        )
    ]
    connections_settings = [
        await wait_for(
            i.get_settings(),
        )
        for i in connections
    ]
    return [
        settings['802-11-wireless']['ssid'][1].decode('utf-8')
        for settings in connections_settings
        if '802-11-wireless' in settings
    ]


async def add_wireless_connection(
    ssid: str,
    password: str,
    type: WiFiType,
    *,
    hidden: bool | None = False,
) -> None:
    wifi_device = await get_wifi_device()
    if not wifi_device:
        return

    access_points = [
        (
            access_point,
            await wait_for(
                access_point.ssid,
            ),
        )
        for access_point in await get_access_points()
    ]
    access_point = next(
        (
            access_point
            for access_point, ssid_ in access_points
            if ssid_.decode('utf8') == ssid
        ),
        None,
    )

    if not access_point:
        return

    if type == WiFiType.nopass:
        security = {
            'key-mgmt': ('s', 'none'),
            'auth-alg': ('s', 'open'),
        }
    elif type == WiFiType.WEP:
        security = {
            'key-mgmt': ('s', 'none'),
            'auth-alg': ('s', 'open'),
            'psk': ('s', password),
        }
    elif type in (WiFiType.WPA, WiFiType.WPA2):
        security = {
            'key-mgmt': ('s', 'wpa-psk'),
            'auth-alg': ('s', 'open'),
            'psk': ('s', password),
        }

    properties: NetworkManagerConnectionProperties = {
        'connection': {
            'id': ('s', ssid),
            'uuid': ('s', str(uuid.uuid4())),
            'type': ('s', '802-11-wireless'),
            'autoconnect': ('b', True),
        },
        '802-11-wireless': {
            'mode': ('s', 'infrastructure'),
            'security': ('s', '802-11-wireless-security'),
            'ssid': ('ay', ssid.encode('utf-8')),
            'hidden': ('b', hidden),
        },
        '802-11-wireless-security': security,
        'ipv4': {'method': ('s', 'auto')},
        'ipv6': {'method': ('s', 'auto')},
    }

    network_manager = NetworkManager(get_system_bus())
    await wait_for(
        network_manager.add_and_activate_connection(
            properties,
            wifi_device._dbus.object_path,  # noqa: SLF001
            access_point._dbus.object_path,  # noqa: SLF001
        ),
    )


async def connect_wireless_connection(ssid: str) -> None:
    wifi_device = await get_wifi_device()

    if not wifi_device:
        return

    network_manager_settings = NetworkManagerSettings(get_system_bus())
    connections = [
        NetworkConnectionSettings(path, get_system_bus())
        for path in await wait_for(
            network_manager_settings.connections,
        )
    ]
    connections_settings = [
        (
            await wait_for(
                connection.get_settings(),
            ),
            connection._dbus.object_path,  # noqa: SLF001
        )
        for connection in connections
    ]
    desired_connection = next(
        (
            path
            for settings, path in connections_settings
            if '802-11-wireless' in settings
            and settings['802-11-wireless']['ssid'][1].decode('utf-8') == ssid
        ),
        None,
    )

    if not desired_connection:
        return

    network_manager = NetworkManager(get_system_bus())
    await wait_for(
        network_manager.activate_connection(
            desired_connection,
            wifi_device._dbus.object_path,  # noqa: SLF001
        ),
    )


async def disconnect_wireless_connection() -> None:
    wifi_device = await get_wifi_device()
    if not wifi_device:
        return

    network_manager = NetworkManager(get_system_bus())
    await wait_for(
        network_manager.deactivate_connection(
            await wait_for(
                wifi_device.active_connection,
            ),
        ),
    )


async def forget_wireless_connection(ssid: str) -> None:
    network_manager_settings = NetworkManagerSettings(get_system_bus())

    for connection_path in await wait_for(
        network_manager_settings.connections,
    ):
        network_connection_settings = NetworkConnectionSettings(
            connection_path,
            get_system_bus(),
        )
        settings = await wait_for(
            network_connection_settings.get_settings(),
        )
        if (
            '802-11-wireless' in settings
            and settings['802-11-wireless']['ssid'][1].decode('utf-8') == ssid
        ):
            await wait_for(network_connection_settings.delete())
            dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title=f'"{ssid}" Deleted',
                        content=f"""WiFi connection with ssid "{
                        ssid}" was deleted successfully""",
                        display_type=NotificationDisplayType.FLASH,
                        color=DANGER_COLOR,
                        icon='󱛅',
                        chime=Chime.DONE,
                    ),
                ),
            )


async def get_connections() -> list[WiFiConnection]:
    # It is need as this action is not atomic and the active_connection may not be
    # available when active_connection.state is queried
    for _ in range(RETRIES):
        with contextlib.suppress(Exception):
            active_connection = await get_active_connection()
            active_connection_ssid = await get_active_connection_ssid()
            saved_ssids = await get_saved_ssids()
            access_point_ssids = {
                (
                    await wait_for(
                        i.ssid,
                    )
                ).decode('utf-8'): i
                for i in await get_access_points()
            }

            active_connection_state = cast(
                SdBusConnectionState,
                await active_connection.state if active_connection else None,
            )

            state_map = {
                SdBusConnectionState.ACTIVATED: ConnectionState.CONNECTED,
                SdBusConnectionState.ACTIVATING: ConnectionState.CONNECTING,
                SdBusConnectionState.DEACTIVATED: ConnectionState.DISCONNECTED,
                SdBusConnectionState.DEACTIVATING: ConnectionState.DISCONNECTED,
                SdBusConnectionState.UNKNOWN: ConnectionState.UNKNOWN,
            }

            return [
                WiFiConnection(
                    ssid=ssid,
                    signal_strength=await wait_for(
                        access_point_ssids[ssid].strength,
                    )
                    if ssid in access_point_ssids
                    else 0,
                    state=state_map[active_connection_state]
                    if active_connection_ssid == ssid
                    else ConnectionState.DISCONNECTED,
                )
                for ssid in saved_ssids
            ]
        await asyncio.sleep(0.5)
    return []



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

from typing import Sequence

from ubo_app.logging import logger
from ubo_app.store import dispatch
from ubo_app.store.services.rgb_ring import RgbRingSetIsConnectedAction
from ubo_app.utils.server import send_command


class RgbRingClient:
    """The instances of this class send commands to the LED manager.

    The LED Manager runs with root privileges due to hardware DMA security constraints.
    The commands are sent through a socket connection to the LED manager.

    The LED manager is a Python script that runs as a daemon. The LED manager
    is a daemon because it uses DMA to control the LEDs. This is a hardware security
    constraint.

    The LED client is a Python script that runs as a non-root user. The LED client
    function is to serialize LED commands and send them to the LED manager in a secure
    manner.
    """

    async def send(self: RgbRingClient, cmd: Sequence[str]) -> None:
        try:
            await send_command(' '.join(['led', *cmd]))
            dispatch(RgbRingSetIsConnectedAction(is_connected=True))
        except Exception as exception:  # noqa: BLE001
            dispatch(RgbRingSetIsConnectedAction(is_connected=False))
            logger.error('Unable to connect to the socket', exc_info=exception)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from debouncer import DebounceOptions, debounce
from wifi_manager import (
    get_connections,
    get_wifi_device,
    get_wifi_device_state,
    request_scan,
)

from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.wifi import (
    ConnectionState,
    WiFiUpdateAction,
    WiFiUpdateRequestEvent,
)
from ubo_app.utils.async_ import create_task


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=False, time_window=2),
)
async def update_wifi_list(_: WiFiUpdateRequestEvent | None = None) -> None:
    connections = await get_connections()

    dispatch(
        WiFiUpdateAction(
            connections=connections,
            state=await get_wifi_device_state(),
            current_connection=next(
                (
                    connection
                    for connection in connections
                    if connection.state is ConnectionState.CONNECTED
                ),
                None,
            ),
        ),
    )


async def setup_listeners() -> None:
    wifi_device = await get_wifi_device()
    if not wifi_device:
        return

    async for _ in wifi_device.properties_changed:
        create_task(update_wifi_list())


def init_service() -> None:
    from pages import main

    create_task(update_wifi_list())
    create_task(setup_listeners())

    dispatch(RegisterSettingAppAction(menu_item=main.WiFiMainMenu))

    subscribe_event(WiFiUpdateRequestEvent, lambda: create_task(request_scan()))



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from enum import StrEnum, auto
from typing import TYPE_CHECKING, Generic, Sequence, TypeVar

from immutable import Immutable
from redux import BaseAction

if TYPE_CHECKING:
    from datetime import datetime


class SensorsAction(BaseAction): ...


Primitive = int | float | str | bool
SensorType = TypeVar('SensorType', bound=Primitive | Sequence[Primitive])


class Sensor(StrEnum):
    TEMPERATURE = auto()
    LIGHT = auto()


class SensorsReportReadingAction(SensorsAction, Generic[SensorType]):
    sensor: Sensor
    reading: SensorType
    timestamp: datetime


class SensorState(Immutable, Generic[SensorType]):
    value: SensorType | None = None


class SensorsState(Immutable):
    temperature: SensorState[float] = SensorState(value=None)
    light: SensorState[float] = SensorState(value=None)



"""Compatibility layer for different environments."""


def setup() -> None:
    """Set up for different environments."""
    from ubo_app.utils import IS_RPI

    if not IS_RPI:
        import sys

        from ubo_app.utils.fake import Fake

        sys.modules['alsaaudio'] = Fake()
        sys.modules['pulsectl'] = Fake()
        sys.modules['pyaudio'] = Fake()
        sys.modules['sdbus'] = Fake()
        sys.modules['sdbus_async'] = Fake()
        sys.modules['sdbus_async.networkmanager'] = Fake()
        sys.modules['sdbus_async.networkmanager.enums'] = Fake()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from threading import current_thread
from typing import TYPE_CHECKING, Awaitable, Callable, TypeVarTuple, Unpack

from typing_extensions import TypeVar

if TYPE_CHECKING:
    from asyncio import Future, Handle

    from redux.basic_types import TaskCreatorCallback


background_tasks: set[Handle] = set()


def create_task(
    awaitable: Awaitable,
    callback: TaskCreatorCallback | None = None,
) -> Handle:
    async def wrapper() -> None:
        from ubo_app.load_services import UboServiceThread
        from ubo_app.logging import get_logger

        logger = get_logger('ubo-app')

        if awaitable is None:
            return
        try:
            logger.verbose(
                'Starting task',
                extra={
                    'awaitable': awaitable,
                    **(
                        {'ubo_service': current_thread().name}
                        if isinstance(current_thread(), UboServiceThread)
                        else {}
                    ),
                },
            )
            await awaitable
        except Exception:
            thread = current_thread()
            logger.exception(
                'Task failed',
                extra={
                    'awaitable': awaitable,
                    **(
                        {'ubo_service': thread.path.as_posix()}
                        if isinstance(thread, UboServiceThread)
                        else {}
                    ),
                },
            )

    import ubo_app.utils.loop

    handle = ubo_app.utils.loop._create_task(wrapper(), callback)  # noqa: SLF001
    background_tasks.add(handle)
    return handle


T = TypeVar('T', infer_variance=True)
Ts = TypeVarTuple('Ts')


def run_in_executor(
    executor: object,
    task: Callable[[Unpack[Ts]], T],
    *args: Unpack[Ts],
) -> Future[T]:
    import ubo_app.utils.loop

    return ubo_app.utils.loop._run_in_executor(executor, task, *args)  # noqa: SLF001



"""Entry point for the bootstrap script."""

import sys

from ubo_app.system.bootstrap import bootstrap


def main() -> None:
    """Run the bootstrap script."""
    bootstrap(
        with_docker='--with-docker' in sys.argv,
        for_packer='--for-packer' in sys.argv,
    )
    sys.stdout.write('Bootstrap completed.\n')
    sys.stdout.flush()
    sys.exit(0)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Sequence

from kivy.clock import Clock
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.stencilview import StencilView
from kivy.uix.widget import Widget
from redux import AutorunOptions
from ubo_gui.app import UboApp

from ubo_app.store import autorun

if TYPE_CHECKING:
    from ubo_app.store.status_icons import IconState


class MenuAppFooter(UboApp):
    def set_temperature_value(self: MenuAppFooter, value: float | None = None) -> None:
        if value is None:
            self.temperature.text = '-'
        else:
            self.temperature.text = f'{value:0.1f}󰔄'

    @cached_property
    def temperature_widget(self: MenuAppFooter) -> BoxLayout:
        layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
            size_hint=(None, 1),
        )

        self.temperature = Label(font_size=dp(14), size_hint=(None, 1), valign='middle')
        self.temperature.bind(
            texture_size=lambda temperature, texture_size: setattr(
                temperature,
                'width',
                texture_size[0],
            )
            or setattr(layout, 'width', temperature.width + dp(12)),
        )

        autorun(
            lambda state: state.sensors.temperature.value,
            options=AutorunOptions(keep_ref=False),
        )(self.set_temperature_value)

        icon = Label(
            text='',
            color='#ffffff',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(12),
        )
        layout.add_widget(icon)
        layout.add_widget(self.temperature)

        return layout

    def set_light_value(self: MenuAppFooter, value: float | None = None) -> None:
        if value is None:
            self.light.color = (0.5, 0, 0, 1)
        else:
            v = min(value, 140) / 140
            self.light.color = (1, 1, 1, v)

    @cached_property
    def light_widget(self: MenuAppFooter) -> Label:
        self.light = Label(
            text='󱩎',
            color='#ffffff',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(16),
        )
        self.light.bind(
            texture_size=lambda light, texture_size: setattr(
                light,
                'width',
                texture_size[0],
            ),
        )

        autorun(
            lambda state: state.sensors.light.value,
            options=AutorunOptions(keep_ref=False),
        )(self.set_light_value)

        return self.light

    @cached_property
    def clock_widget(self: MenuAppFooter) -> Label:
        clock = Label(font_size=dp(20), size_hint=(None, 1), valign='middle')
        clock.bind(
            texture_size=lambda clock, texture_size: setattr(
                clock,
                'width',
                texture_size[0],
            ),
        )
        local_timzone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

        def now() -> datetime.datetime:
            return datetime.datetime.now(local_timzone)

        def set_value(_: int | None = None) -> None:
            clock.text = now().strftime('%H:%M')

        def update(_: int | None = None) -> None:
            set_value()
            now_ = now()
            Clock.schedule_once(
                update,
                60 - now_.second - now_.microsecond / 1000000 + 0.05,
            )

        update()

        return clock

    def render_icons(
        self: MenuAppFooter,
        selector_result: Sequence[IconState],
    ) -> None:
        icons = selector_result
        self.icons_layout.clear_widgets()
        for icon in list(reversed(icons))[:4]:
            label = Label(
                text=icon.symbol,
                color=icon.color,
                font_size=dp(20),
                font_features='fill=0',
                size_hint=(None, 1),
                width=dp(22),
            )
            self.icons_layout.add_widget(label)
        self.icons_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.icons_layout.bind(minimum_width=self.icons_layout.setter('width'))

    def handle_depth_change(self: MenuAppFooter, _: Sequence[str]) -> None:
        is_fullscreen = False
        if not is_fullscreen:
            if self.normal_footer_layout in self.footer_layout.children:
                self.footer_layout.remove_widget(self.normal_footer_layout)
                self.footer_layout.add_widget(self.home_footer_layout)
        elif self.home_footer_layout in self.footer_layout.children:
            self.footer_layout.remove_widget(self.home_footer_layout)
            self.footer_layout.add_widget(self.normal_footer_layout)

    def set_icons_layout_x(self: MenuAppFooter, *_: list[Any]) -> None:
        self.icons_layout.x = (
            self.icons_widget.x + self.icons_widget.width - self.icons_layout.width
        )

    @cached_property
    def footer(self: MenuAppFooter) -> Widget | None:
        self.footer_layout = BoxLayout()

        self.normal_footer_layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
        )
        self.normal_footer_layout.add_widget(
            Label(
                text='',
                font_size=dp(20),
                size_hint=(None, 1),
            ),
        )
        self.normal_footer_layout.add_widget(Widget(size_hint=(1, 1)))

        self.home_footer_layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
        )

        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.clock_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.temperature_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.light_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))

        self.icons_widget = StencilView(size_hint=(1, 1))
        self.home_footer_layout.add_widget(self.icons_widget)

        self.icons_layout = BoxLayout(
            orientation='horizontal',
            spacing=dp(4),
            padding=0,
        )
        self.icons_widget.add_widget(self.icons_layout)

        self.icons_widget.bind(
            width=self.set_icons_layout_x,
            x=self.set_icons_layout_x,
            height=self.icons_layout.setter('height'),
            y=self.icons_layout.setter('y'),
        )
        self.icons_layout.bind(
            width=self.set_icons_layout_x,
            x=self.set_icons_layout_x,
        )

        autorun(
            lambda state: state.status_icons.icons,
            options=AutorunOptions(keep_ref=False),
        )(self.render_icons)

        autorun(
            lambda state: state.main.path,
            options=AutorunOptions(keep_ref=False),
        )(self.handle_depth_change)

        self.footer_layout.add_widget(self.home_footer_layout)

        return self.footer_layout



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import asyncio
import socket
from collections import defaultdict
from typing import Sequence

import psutil
from constants import INTERNET_STATE_ICON_ID, INTERNET_STATE_ICON_PRIORITY
from ubo_gui.menu.types import HeadlessMenu, Item, SubMenuItem

from ubo_app.store import autorun, dispatch, subscribe_event
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.ip import (
    IpNetworkInterface,
    IpUpdateAction,
    IpUpdateRequestAction,
    IpUpdateRequestEvent,
)
from ubo_app.store.status_icons import StatusIconsRegisterAction


@autorun(lambda state: state.ip.interfaces)
def get_ip_addresses(interfaces: Sequence[IpNetworkInterface]) -> list[SubMenuItem]:
    if not interfaces:
        return []
    return [
        SubMenuItem(
            label=interface.name,
            icon='󰈀'
            if interface.name.startswith('eth')
            else ''
            if interface.name.startswith('wlan')
            else '󰕇'
            if interface.name.startswith('lo')
            else '󰛳',
            sub_menu=HeadlessMenu(
                title=f'IP Addresses - {interface.name}',
                items=[
                    Item(
                        label=ip_address,
                        icon='󰩠',
                    )
                    for ip_address in interface.ip_addresses
                ],
                placeholder='No IP addresses',
            ),
        )
        for interface in interfaces
    ]


def load_ip_addresses() -> None:
    ip_addresses_by_interface = defaultdict(list)
    for interface_name, ip_addresses in psutil.net_if_addrs().items():
        for address in ip_addresses:
            if address.family == socket.AddressFamily.AF_INET:
                ip_addresses_by_interface[interface_name].append(address.address)

    dispatch(
        IpUpdateAction(
            interfaces=[
                IpNetworkInterface(name=interface_name, ip_addresses=ip_addresses)
                for interface_name, ip_addresses in ip_addresses_by_interface.items()
            ],
        ),
    )


def is_connected() -> bool:
    try:
        with socket.create_connection(('1.1.1.1', 53), timeout=2):
            return True
    except OSError:
        return False


async def check_connection() -> bool:
    while True:
        await asyncio.sleep(1)
        if is_connected():
            dispatch(
                IpUpdateRequestAction(),
                StatusIconsRegisterAction(
                    icon='󰖟',
                    priority=INTERNET_STATE_ICON_PRIORITY,
                    id=INTERNET_STATE_ICON_ID,
                ),
            )
        else:
            dispatch(
                IpUpdateRequestAction(),
                StatusIconsRegisterAction(
                    icon='󰪎',
                    priority=INTERNET_STATE_ICON_PRIORITY,
                    id=INTERNET_STATE_ICON_ID,
                ),
            )


IpMainMenu = SubMenuItem(
    label='IP Addresses',
    icon='󰩟',
    sub_menu=HeadlessMenu(
        title='IP Addresses',
        items=get_ip_addresses,
        placeholder='No IP addresses',
    ),
)


async def init_service() -> None:
    dispatch(RegisterSettingAppAction(menu_item=IpMainMenu))

    subscribe_event(
        IpUpdateRequestEvent,
        load_ip_addresses,
    )
    load_ip_addresses()
    await check_connection()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import sys
import threading
import traceback
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


def setup_sentry() -> None:  # pragma: no cover
    import os
    from asyncio import CancelledError

    import sentry_sdk

    if 'SENTRY_DSN' in os.environ:
        sentry_sdk.init(
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
            ignore_errors=[KeyboardInterrupt, CancelledError],
        )


def get_all_thread_stacks() -> dict[str, list[str]]:
    id_to_name = {th.ident: th.name for th in threading.enumerate()}
    thread_stacks = {}
    for thread_id, frame in sys._current_frames().items():  # noqa: SLF001
        thread_stacks[id_to_name.get(thread_id, f'unknown-{thread_id}')] = (
            traceback.format_stack(frame)
        )
    return thread_stacks


def global_exception_handler(
    exception_type: type[BaseException],
    exception_value: BaseException,
    exception_traceback: TracebackType,
) -> None:
    from ubo_app.logging import logger

    error_message = ''.join(
        traceback.format_exception(
            exception_type,
            exception_value,
            exception_traceback,
        ),
    )
    threads_info = get_all_thread_stacks()

    logger.error(
        f"""Uncaught exception: {exception_type}: {exception_value}
{error_message}""",
    )
    logger.debug(
        f"""Uncaught exception: {exception_type}: {exception_value}
{error_message}""",
        extra={'threads': threads_info},
    )


def thread_exception_handler(args: threading.ExceptHookArgs) -> None:
    from ubo_app.logging import logger

    error_message = ''.join(
        traceback.format_exception(*args[:3]),
    )
    threads_info = get_all_thread_stacks()

    exception_type, exception_value, _, thread = args

    logger.error(
        f"""Uncaught exception in thread {thread}: {exception_type}: {exception_value}
{error_message}""",
    )
    logger.debug(
        f"""Uncaught exception in thread {thread}: {exception_type}: {exception_value}
{error_message}""",
        extra={'threads': threads_info, 'thread': thread},
    )


def setup_error_handling() -> None:
    setup_sentry()
    threading.excepthook = thread_exception_handler
    sys.excepthook = global_exception_handler



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.services.rgb_ring import RgbRingCommandEvent, RgbRingPulseAction


def init_service() -> None:
    from rgb_ring_client import RgbRingClient

    rgb_ring_client = RgbRingClient()

    async def handle_rgb_ring_command(event: RgbRingCommandEvent) -> None:
        await rgb_ring_client.send(event.command)

    subscribe_event(RgbRingCommandEvent, handle_rgb_ring_command)

    dispatch(RgbRingPulseAction(repetitions=2, wait=180))



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_app.services import register


def setup() -> None:
    from setup import init_service

    init_service()


register(
    service_id='keypad',
    label='Keypad',
    setup=setup,
)



"""provides a function to install and start Docker on the host machine."""

from __future__ import annotations

import subprocess
from pathlib import Path

from ubo_app.constants import DOCKER_INSTALLATION_LOCK_FILE, USERNAME


def docker_handler(command: str) -> None:
    """Install and start Docker on the host machine."""
    Path.touch(
        DOCKER_INSTALLATION_LOCK_FILE,
        mode=0o666,
        exist_ok=True,
    )
    try:
        if command == 'install':
            subprocess.run(
                Path(__file__).parent.parent.joinpath('install_docker.sh'),  # noqa: S603
                env={'USERNAME': USERNAME},
                check=False,
            )
        elif command == 'start':
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'systemctl',
                    'start',
                    'docker.socket',
                ],
                check=False,
            )
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'systemctl',
                    'start',
                    'docker.service',
                ],
                check=False,
            )
        elif command == 'stop':
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'systemctl',
                    'stop',
                    'docker.socket',
                ],
                check=False,
            )
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'systemctl',
                    'stop',
                    'docker.service',
                ],
                check=False,
            )
    finally:
        DOCKER_INSTALLATION_LOCK_FILE.unlink(missing_ok=True)



# pyright: reportMissingImports=false
# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

import asyncio
from threading import current_thread
from typing import TYPE_CHECKING, Any, Coroutine, TypeVar

from ubo_app.store.services.ethernet import GlobalEthernetState

if TYPE_CHECKING:
    from asyncio.tasks import _FutureLike


T = TypeVar('T')


def wait_for(task: _FutureLike[T]) -> Coroutine[Any, Any, T]:
    return asyncio.wait_for(task, timeout=10.0)


from sdbus import SdBus, sd_bus_open_system, set_default_bus  # noqa: E402
from sdbus_async.networkmanager import (  # noqa: E402
    DeviceState,
    NetworkDeviceGeneric,
    NetworkManager,
)
from sdbus_async.networkmanager.enums import DeviceType  # noqa: E402

system_buses = {}


def get_system_bus() -> SdBus:
    thread = current_thread()
    if thread not in system_buses:
        system_buses[thread] = sd_bus_open_system()
    set_default_bus(system_buses[thread])
    return system_buses[thread]


async def get_ethernet_device() -> NetworkDeviceGeneric | None:
    network_manager = NetworkManager(get_system_bus())
    devices_paths = await wait_for(
        network_manager.get_devices(),
    )
    for device_path in devices_paths:
        generic_device = NetworkDeviceGeneric(device_path, get_system_bus())
        if (
            await wait_for(
                generic_device.device_type,
            )
            == DeviceType.ETHERNET
        ):
            return generic_device
    return None


async def get_ethernet_device_state() -> GlobalEthernetState:
    ethernet_device = await get_ethernet_device()
    if ethernet_device is None:
        return GlobalEthernetState.UNKNOWN

    state = await ethernet_device.state
    if state is DeviceState.UNKNOWN:
        return GlobalEthernetState.UNKNOWN
    if state in (
        DeviceState.DISCONNECTED,
        DeviceState.UNMANAGED,
        DeviceState.UNAVAILABLE,
        DeviceState.FAILED,
    ):
        return GlobalEthernetState.DISCONNECTED
    if state in (DeviceState.NEED_AUTH,):
        return GlobalEthernetState.NEEDS_ATTENTION
    if state in (
        DeviceState.DEACTIVATING,
        DeviceState.PREPARE,
        DeviceState.CONFIG,
        DeviceState.IP_CONFIG,
        DeviceState.IP_CHECK,
        DeviceState.SECONDARIES,
    ):
        return GlobalEthernetState.PENDING
    if state == DeviceState.ACTIVATED:
        return GlobalEthernetState.CONNECTED

    return GlobalEthernetState.UNKNOWN



"""SSH service module."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Sequence

from ubo_gui.menu.types import (
    ActionItem,
    ApplicationItem,
    HeadedMenu,
    HeadlessMenu,
    Item,
    Menu,
    SubMenuItem,
)
from ubo_gui.prompt import PromptWidget

from ubo_app.store import autorun, dispatch
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.ssh import SSHClearEnabledStateAction, SSHUpdateStateAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.monitor_unit import is_unit_active, is_unit_enabled, monitor_unit
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from ubo_app.store.services.ssh import SSHState


class ClearTemporaryUsersPrompt(PromptWidget):
    """Prompt to clear all temporary users."""

    def first_option_callback(self: ClearTemporaryUsersPrompt) -> None:
        """Clear all temporary users."""
        self.dispatch('on_close')

    def second_option_callback(self: ClearTemporaryUsersPrompt) -> None:
        """Close the prompt."""

        async def act() -> None:
            await send_command('service ssh clear_all_temporary_accounts')
            self.dispatch('on_close')
            dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        title='All SSH Accounts Removed',
                        content='All SSH accounts have been removed.',
                        importance=Importance.MEDIUM,
                        icon='󰣀',
                        display_type=NotificationDisplayType.FLASH,
                    ),
                ),
            )

        create_task(act())

    def __init__(self: ClearTemporaryUsersPrompt, **kwargs: object) -> None:
        """Initialize the prompt."""
        super().__init__(**kwargs, items=None)
        self.prompt = 'Clear temporary ssh users?'
        self.icon = '󱈋'
        self.first_option_label = ''
        self.first_option_background_color = 'black'
        self.first_option_is_short = False
        self.second_option_label = 'Delete'
        self.second_option_icon = '󱈊'
        self.second_option_is_short = False


def create_ssh_account() -> None:
    """Create a temporary SSH account."""

    async def act() -> None:
        result = await send_command(
            'service ssh create_temporary_ssh_account',
            has_output=True,
        )
        username, password = result.split(':')
        dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Temporary SSH Account Created',
                    content=f"""Username: {username}
Password: {password}
Make sure to delete it after use. Note that in order to make things work for you, we \
had to make sure password authentication for ssh server is enabled, you may want to \
disable it later. Clearing all temporary users will disable password authentication \
too.""",
                    importance=Importance.MEDIUM,
                    icon='󰣀',
                    display_type=NotificationDisplayType.STICKY,
                ),
            ),
        )

    create_task(act())


def start_ssh_service() -> None:
    """Start the SSH service."""

    async def act() -> None:
        await send_command('service ssh start')

    create_task(act())


def stop_ssh_service() -> None:
    """Stop the SSH service."""

    async def act() -> None:
        await send_command('service ssh stop')

    create_task(act())


def enable_ssh_service() -> None:
    """Enable the SSH service."""

    async def act() -> None:
        dispatch(SSHClearEnabledStateAction())
        await send_command('service ssh enable')
        await asyncio.sleep(5)
        await check_is_ssh_enabled()

    create_task(act())


def disable_ssh_service() -> None:
    """Disable the SSH service."""

    async def act() -> None:
        dispatch(SSHClearEnabledStateAction())
        await send_command('service ssh disable')
        await asyncio.sleep(5)
        await check_is_ssh_enabled()

    create_task(act())


@autorun(lambda state: state.ssh)
def ssh_items(state: SSHState) -> Sequence[Item]:
    """Get the SSH menu items."""
    return [
        SubMenuItem(
            label='Create account',
            icon='',
            sub_menu=HeadedMenu(
                title='SSH Setup',
                heading='Create an SSH account',
                sub_heading='This will create a temporary SSH account,'
                ' you should delete it after use.',
                items=[
                    ActionItem(
                        label='Create',
                        icon='󰗹',
                        action=create_ssh_account,
                    ),
                ],
            ),
        ),
        ApplicationItem(
            label='Remove all users',
            icon='󱈊',
            application=ClearTemporaryUsersPrompt,
        ),
        ActionItem(
            label='Stop' if state.is_active else 'Start',
            icon='󰓛' if state.is_active else '󰐊',
            action=stop_ssh_service if state.is_active else start_ssh_service,
        ),
        Item(
            label='...',
            icon='',
        )
        if state.is_enabled is None
        else ActionItem(
            label='Disable',
            icon='[color=#008000]󰯄[/color]',
            action=disable_ssh_service,
        )
        if state.is_enabled
        else ActionItem(
            label='Enable',
            icon='[color=#ffff00]󰯅[/color]',
            action=enable_ssh_service,
        ),
    ]


@autorun(lambda state: state.ssh)
def ssh_icon(state: SSHState) -> str:
    """Get the SSH icon."""
    return '[color=#008000]󰪥[/color]' if state.is_active else '[color=#ffff00]󰝦[/color]'


@autorun(lambda state: state.ssh)
def ssh_title(_: SSHState) -> str:
    """Get the SSH title."""
    return ssh_icon() + ' SSH'


async def check_is_ssh_active() -> None:
    """Check if the SSH service is active."""
    if await is_unit_active('sshd'):
        dispatch(SSHUpdateStateAction(is_enabled=True))
    else:
        dispatch(SSHUpdateStateAction(is_enabled=False))


async def check_is_ssh_enabled() -> None:
    """Check if the SSH service is enabled."""
    if await is_unit_enabled('sshd'):
        dispatch(SSHUpdateStateAction(is_enabled=True))
    else:
        dispatch(SSHUpdateStateAction(is_enabled=False))


def open_ssh_menu() -> Menu:
    """Open the SSH menu."""
    create_task(
        asyncio.gather(
            check_is_ssh_active(),
            check_is_ssh_enabled(),
        ),
    )

    return HeadlessMenu(
        title=ssh_title,
        items=ssh_items,
    )


def init_service() -> None:
    """Initialize the SSH service."""
    dispatch(
        RegisterSettingAppAction(
            menu_item=ActionItem(
                label='SSH',
                icon=ssh_icon,
                action=open_ssh_menu,
            ),
        ),
    )

    create_task(
        asyncio.gather(
            check_is_ssh_active(),
            check_is_ssh_enabled(),
            monitor_unit(
                'ssh.service',
                lambda status: dispatch(
                    SSHUpdateStateAction(
                        is_active=status in ('active', 'activating', 'reloading'),
                    ),
                ),
            ),
        ),
    )



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107, N999
from __future__ import annotations

import pathlib
from distutils.util import strtobool
from typing import Sequence, cast

from kivy.clock import mainthread
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty
from ubo_gui.constants import SUCCESS_COLOR
from ubo_gui.menu.types import ActionItem, Item
from ubo_gui.page import PageWidget
from wifi_manager import add_wireless_connection

from ubo_app.logging import logger
from ubo_app.store import dispatch
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.sound import SoundPlayChimeAction
from ubo_app.store.services.wifi import WiFiType, WiFiUpdateRequestAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.qrcode import qrcode_input

# Regular expression pattern
# WIFI:S:<SSID>;T:<WEP|WPA|blank>;P:<PASSWORD>;H:<true|false|blank>;;
BARCODE_PATTERN = (
    r'^WIFI:S:(?P<SSID>[^;]*);(?:T:(?P<Type>(?i:WEP|WPA|WPA2|nopass));)'
    r'?(?:P:(?P<Password>[^;]*);)?(?:H:(?P<Hidden>(?i:true|false));)?;$'
)


class CreateWirelessConnectionPage(PageWidget):
    creating = BooleanProperty(defaultvalue=False)

    def __init__(
        self: CreateWirelessConnectionPage,
        items: Sequence[Item] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(*args, **kwargs, items=items)
        dispatch(SoundPlayChimeAction(name='scan'))

    async def create_wireless_connection(self: CreateWirelessConnectionPage) -> None:
        _, match = await qrcode_input(BARCODE_PATTERN)
        if not match:
            return
        ssid = match.get('SSID')
        if ssid is None:
            return

        password = match.get('Password')
        type = cast(WiFiType, match.get('Type'))
        hidden = strtobool(match.get('Hidden') or 'false') == 1

        if not password:
            logger.warn('Password is required')
            return

        self.creating = True

        await add_wireless_connection(
            ssid=ssid,
            password=password,
            type=type or WiFiType.nopass,
            hidden=hidden,
        )

        logger.info(
            'Wireless connection created',
            extra={
                'ssid': ssid,
                'type': type,
                'hidden': hidden,
            },
        )

        dispatch(
            WiFiUpdateRequestAction(reset=True),
            NotificationsAddAction(
                notification=Notification(
                    title=f'"{ssid}" Added',
                    content=f"""WiFi connection with ssid "{
                    ssid}" was added successfully""",
                    display_type=NotificationDisplayType.FLASH,
                    color=SUCCESS_COLOR,
                    icon='󱛃',
                    chime=Chime.ADD,
                ),
            ),
        )
        mainthread(self.dispatch)('on_close')

    def input_connection_information(self: CreateWirelessConnectionPage) -> None:
        create_task(self.create_wireless_connection())

    def get_item(self: CreateWirelessConnectionPage, index: int) -> ActionItem | None:
        if index == 2:  # noqa: PLR2004
            return ActionItem(
                label='start',
                is_short=True,
                icon='󰄀',
                action=self.input_connection_information,
            )
        return super().get_item(index)


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('create_wireless_connection.kv')
    .resolve()
    .as_posix(),
)



"""Docker image menu."""

from __future__ import annotations

import contextlib
import pathlib
from asyncio import iscoroutine
from typing import Any, Callable, Coroutine, Mapping, cast, overload

import docker
import docker.errors
from docker.models.containers import Container
from docker.models.images import Image
from kivy.lang.builder import Builder
from kivy.properties import ListProperty, NumericProperty, StringProperty
from reducer import IMAGE_IDS, IMAGES
from ubo_gui.menu.types import ActionItem, HeadedMenu, HeadlessMenu, Item, SubMenuItem
from ubo_gui.page import PageWidget

from ubo_app.logging import logger
from ubo_app.store import autorun, dispatch
from ubo_app.store.services.docker import (
    DockerImageSetDockerIdAction,
    DockerImageSetStatusAction,
    DockerState,
    ImageState,
    ImageStatus,
)
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationsAddAction,
)
from ubo_app.utils.async_ import create_task, run_in_executor


def find_container(client: docker.DockerClient, *, image: str) -> Container | None:
    """Find a container."""
    for container in client.containers.list(all=True):
        if not isinstance(container, Container):
            continue

        with contextlib.suppress(docker.errors.DockerException):
            container_image = container.image
            if isinstance(container_image, Image) and image in container_image.tags:
                return container
    return None


def update_container(image_id: str, container: Container) -> None:
    """Update a container's state in store based on its real state."""
    if container.status == 'running':
        logger.debug(
            'Container running image found',
            extra={'image': image_id, 'path': IMAGES[image_id].path},
        )
        dispatch(
            DockerImageSetStatusAction(
                image=image_id,
                status=ImageStatus.RUNNING,
                ports=[
                    f'{i["HostIp"]}:{i["HostPort"]}'
                    for i in container.ports.values()
                    for i in i
                ],
                ip=container.attrs['NetworkSettings']['Networks']['bridge']['IPAddress']
                if container.attrs
                and 'bridge' in container.attrs['NetworkSettings']['Networks']
                else None,
            ),
        )
        return
    logger.debug(
        "Container for the image found, but it's not running",
        extra={'image': image_id, 'path': IMAGES[image_id].path},
    )
    dispatch(
        DockerImageSetStatusAction(
            image=image_id,
            status=ImageStatus.CREATED,
        ),
    )


def _monitor_events(  # noqa: C901
    image_id: str,
    get_docker_id: Callable[[], str],
    docker_client: docker.DockerClient,
) -> None:
    path = IMAGES[image_id].path
    events = docker_client.events(
        decode=True,
        filters={'type': ['image', 'container']},
    )
    for event in events:
        logger.verbose('Docker image event', extra={'event': event})
        if event['Type'] == 'image':
            if event['status'] == 'pull' and event['id'] == path:
                try:
                    image = docker_client.images.get(path)
                    dispatch(
                        DockerImageSetStatusAction(
                            image=image_id,
                            status=ImageStatus.AVAILABLE,
                        ),
                    )
                    if isinstance(image, Image) and image.id:
                        dispatch(
                            DockerImageSetDockerIdAction(
                                image=image_id,
                                docker_id=image.id,
                            ),
                        )
                except docker.errors.DockerException:
                    dispatch(
                        DockerImageSetStatusAction(
                            image=image_id,
                            status=ImageStatus.NOT_AVAILABLE,
                        ),
                    )
            elif event['status'] == 'delete' and event['id'] == get_docker_id():
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.NOT_AVAILABLE,
                    ),
                )
        elif event['Type'] == 'container':
            if event['status'] == 'start' and event['from'] == path:
                container = find_container(docker_client, image=path)
                if container:
                    update_container(image_id, container)
            elif event['status'] == 'die' and event['from'] == path:
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.CREATED,
                    ),
                )
            elif event['status'] == 'destroy' and event['from'] == path:
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.AVAILABLE,
                    ),
                )


def check_container(image_id: str) -> None:
    """Check the container status."""
    path = IMAGES[image_id].path

    def act() -> None:
        logger.debug('Checking image', extra={'image': image_id, 'path': path})
        docker_client = docker.from_env()
        try:
            image = docker_client.images.get(path)
            if not isinstance(image, Image):
                raise docker.errors.ImageNotFound(path)  # noqa: TRY301

            if image.id:
                dispatch(
                    DockerImageSetDockerIdAction(
                        image=image_id,
                        docker_id=image.id,
                    ),
                )
            logger.debug('Image found', extra={'image': image_id, 'path': path})

            container = find_container(docker_client, image=path)
            if container:
                update_container(image_id, container)
                return

            logger.debug(
                'Container running image not found',
                extra={'image': image_id, 'path': path},
            )
            dispatch(
                DockerImageSetStatusAction(
                    image=image_id,
                    status=ImageStatus.AVAILABLE,
                ),
            )
        except docker.errors.ImageNotFound as exception:
            logger.debug(
                'Image not found',
                extra={'image': image_id, 'path': path},
                exc_info=exception,
            )
            dispatch(
                DockerImageSetStatusAction(
                    image=image_id,
                    status=ImageStatus.NOT_AVAILABLE,
                ),
            )
        except docker.errors.DockerException as exception:
            logger.debug(
                'Image error',
                extra={'image': image_id, 'path': path},
                exc_info=exception,
            )
            dispatch(
                DockerImageSetStatusAction(
                    image=image_id,
                    status=ImageStatus.ERROR,
                ),
            )
        finally:

            @autorun(lambda state: getattr(state.docker, image_id).docker_id)
            def get_docker_id(docker_id: str) -> str:
                return docker_id

            _monitor_events(image_id, get_docker_id, docker_client)
            docker_client.close()

    run_in_executor(None, act)


def _fetch_image(image: ImageState) -> None:
    def act() -> None:
        dispatch(
            DockerImageSetStatusAction(
                image=image.id,
                status=ImageStatus.FETCHING,
            ),
        )
        try:
            logger.debug('Fetching image', extra={'image': image.path})
            docker_client = docker.from_env()
            response = docker_client.api.pull(
                image.path,
                stream=True,
                decode=True,
            )
            for line in response:
                dispatch(
                    DockerImageSetStatusAction(
                        image=image.id,
                        status=ImageStatus.FETCHING,
                    ),
                )
                logger.verbose(
                    'Image pull',
                    extra={'image': image.path, 'line': line},
                )
            logger.debug('Image fetched', extra={'image': image.path})
            docker_client.close()
        except docker.errors.DockerException as exception:
            logger.debug(
                'Image error',
                extra={'image': image.path},
                exc_info=exception,
            )
            dispatch(
                DockerImageSetStatusAction(
                    image=image.id,
                    status=ImageStatus.ERROR,
                ),
            )

    run_in_executor(None, act)


def _remove_image(image: ImageState) -> None:
    def act() -> None:
        docker_client = docker.from_env()
        docker_client.images.remove(image.path, force=True)
        docker_client.close()

    run_in_executor(None, act)


@overload
async def _process_str(
    value: str
    | Callable[[], str | Coroutine[Any, Any, str]]
    | Coroutine[Any, Any, str],
) -> str: ...


@overload
async def _process_str(
    value: str
    | Callable[[], str | Coroutine[Any, Any, str | None] | None]
    | Coroutine[Any, Any, str | None]
    | None,
) -> str | None: ...


async def _process_str(
    value: str
    | Callable[[], str | Coroutine[Any, Any, str | None] | None]
    | Coroutine[Any, Any, str | None]
    | None,
) -> str | None:
    if callable(value):
        value = value()
    if iscoroutine(value):
        value = cast(str, await value)
    return cast(str | None, value)


async def _process_environment_variables(image_id: str) -> Mapping[str, str]:
    environment_variables = IMAGES[image_id].environment_vairables or {}
    result: dict[str, str] = {}

    for key in environment_variables:
        result[key] = await _process_str(environment_variables[key])

    return result


@autorun(lambda state: state.docker)
def _run_container_generator(docker_state: DockerState) -> Callable[[ImageState], None]:
    def run_container(image: ImageState) -> None:
        async def act() -> None:
            docker_client = docker.from_env()
            container = find_container(docker_client, image=image.path)
            if container:
                if container.status != 'running':
                    container.start()
            else:
                hosts = {}
                for key, value in IMAGES[image.id].hosts.items():
                    if not hasattr(docker_state, value):
                        dispatch(
                            NotificationsAddAction(
                                notification=Notification(
                                    title='Dependency error',
                                    content=f'Container "{value}" is not loaded',
                                    importance=Importance.MEDIUM,
                                ),
                            ),
                        )
                        return
                    if not getattr(docker_state, value).container_ip:
                        dispatch(
                            NotificationsAddAction(
                                notification=Notification(
                                    title='Dependency error',
                                    content=f'Container "{value}" does not have an IP'
                                    ' address',
                                    importance=Importance.MEDIUM,
                                ),
                            ),
                        )
                        return
                    if hasattr(docker_state, value):
                        hosts[key] = getattr(docker_state, value).container_ip
                    else:
                        hosts[key] = value

                docker_client.containers.run(
                    image.path,
                    hostname=image.id,
                    publish_all_ports=True,
                    detach=True,
                    volumes=IMAGES[image.id].volumes,
                    ports=IMAGES[image.id].ports,
                    network_mode=IMAGES[image.id].network_mode,
                    environment=await _process_environment_variables(image.id),
                    extra_hosts=hosts,
                    restart_policy={'Name': 'always'},
                    command=await _process_str(IMAGES[image.id].command),
                )
            docker_client.close()

        create_task(act())

    return run_container


def _stop_container(image: ImageState) -> None:
    def act() -> None:
        docker_client = docker.from_env()
        container = find_container(docker_client, image=image.path)
        if container and container.status != 'exited':
            container.stop()
        docker_client.close()

    run_in_executor(None, act)


def _remove_container(image: ImageState) -> None:
    def act() -> None:
        docker_client = docker.from_env()
        container = find_container(docker_client, image=image.path)
        if container:
            container.remove(v=True, force=True)
        docker_client.close()

    run_in_executor(None, act)


class DockerQRCode(PageWidget):
    """QR code for the container's url (ip and port)."""

    ips: list[str] = ListProperty()
    port: str = StringProperty()
    index: int = NumericProperty(0)

    def go_down(self: DockerQRCode) -> None:
        """Go down."""
        self.index = (self.index + 1) % len(self.ips)
        self.ids.slider.animated_value = len(self.ips) - 1 - self.index

    def go_up(self: DockerQRCode) -> None:
        """Go up."""
        self.index = (self.index - 1) % len(self.ips)
        self.ids.slider.animated_value = len(self.ips) - 1 - self.index


def image_menu(
    image: ImageState,
) -> HeadedMenu:
    """Get the menu for the docker image."""
    items: list[Item] = []

    def open_qrcode(port: str) -> Callable[[], PageWidget]:
        def action() -> PageWidget:
            return DockerQRCode(ips=image.ip_addresses, port=port)

        return action

    if image.status == ImageStatus.NOT_AVAILABLE:
        items.append(
            ActionItem(
                label='Fetch',
                icon='󰇚',
                action=lambda: _fetch_image(image),
            ),
        )
    elif image.status == ImageStatus.FETCHING:
        items.append(
            ActionItem(
                label='Stop',
                icon='󰓛',
                action=lambda: _remove_image(image),
            ),
        )
    elif image.status == ImageStatus.AVAILABLE:
        items.extend(
            [
                ActionItem(
                    label='Start',
                    icon='󰐊',
                    action=lambda: _run_container_generator()(image),
                ),
                ActionItem(
                    label='Remove image',
                    icon='󰆴',
                    action=lambda: _remove_image(image),
                ),
            ],
        )
    elif image.status == ImageStatus.CREATED:
        items.extend(
            [
                ActionItem(
                    label='Start',
                    icon='󰐊',
                    action=lambda: _run_container_generator()(image),
                ),
                ActionItem(
                    label='Remove container',
                    icon='󰆴',
                    action=lambda: _remove_container(image),
                ),
            ],
        )
    elif image.status == ImageStatus.RUNNING:
        items.append(
            ActionItem(
                label='Stop',
                icon='󰓛',
                action=lambda: _stop_container(image),
            ),
        )
        items.append(
            SubMenuItem(
                label='Ports',
                icon='󰙜',
                sub_menu=HeadlessMenu(
                    title='Ports',
                    items=[
                        ActionItem(
                            label=port,
                            icon='󰙜',
                            action=open_qrcode(port.split(':')[-1]),
                        )
                        if port.startswith('0.0.0.0')  # noqa: S104
                        else Item(label=port, icon='󰙜')
                        for port in image.ports
                    ],
                    placeholder='No ports',
                ),
            ),
        )

    return HeadedMenu(
        title=f'Docker - {image.label}',
        heading=image.label,
        sub_heading={
            ImageStatus.NOT_AVAILABLE: 'Image needs to be fetched',
            ImageStatus.FETCHING: 'Image is being fetched',
            ImageStatus.AVAILABLE: 'Image is ready but container is not running',
            ImageStatus.CREATED: 'Container is created but not running',
            ImageStatus.RUNNING: IMAGES[image.id].note or 'Container is running',
            ImageStatus.ERROR: 'Image has an error, please check the logs',
        }[image.status],
        items=items,
    )


def image_menu_generator(image_id: str) -> Callable[[], Callable[[], HeadedMenu]]:
    """Get the menu items for the Docker service."""
    _image_menu = autorun(lambda state: getattr(state.docker, image_id))(image_menu)

    def image_action() -> Callable[[], HeadedMenu]:
        """Get the menu items for the Docker service."""
        check_container(image_id)

        return _image_menu

    return image_action


image_menus = {image_id: image_menu_generator(image_id) for image_id in IMAGE_IDS}
Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('docker_qrcode.kv').resolve().as_posix(),
)



"""provides a function to interact with system services."""

from __future__ import annotations

import subprocess
from pathlib import Path

from ubo_app.logging import get_logger


def ssh_handler(command: str) -> str | None:
    """Handle ssh commands."""
    if command == 'create_temporary_ssh_account':
        result = subprocess.run(
            Path(__file__).parent.joinpath('create_temporary_ssh_account.sh'),  # noqa: S603
            check=True,
            text=True,
            stdout=subprocess.PIPE,
        )
        result.check_returncode()
        return result.stdout
    if command == 'clear_all_temporary_accounts':
        subprocess.run(
            Path(__file__).parent.joinpath('clear_all_temporary_accounts.sh'),  # noqa: S603
            check=False,
        )
    if command == 'start':
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'start', 'sshd'],  # noqa: S603
            check=True,
        )
    if command == 'stop':
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'stop', 'sshd'],  # noqa: S603
            check=True,
        )
    if command == 'enable':
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'enable', 'sshd'],  # noqa: S603
            check=True,
        )
    if command == 'disable':
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'disable', 'sshd'],  # noqa: S603
            check=True,
        )
    msg = f'Invalid ssh command "{command}"'
    raise ValueError(msg)


def system_handler(service: str, command: str) -> str | None:
    """Interact with system services."""
    logger = get_logger('system-manager')
    try:
        if service == 'ssh':
            return ssh_handler(command)
        if service == 'lightdm':
            if command == 'start':
                subprocess.run(
                    ['/usr/bin/env', 'systemctl', 'start', 'lightdm'],  # noqa: S603
                    check=True,
                )
            elif command == 'stop':
                subprocess.run(
                    ['/usr/bin/env', 'systemctl', 'stop', 'lightdm'],  # noqa: S603
                    check=True,
                )
            elif command == 'enable':
                subprocess.run(
                    ['/usr/bin/env', 'systemctl', 'enable', 'lightdm'],  # noqa: S603
                    check=True,
                )
            elif command == 'disable':
                subprocess.run(
                    ['/usr/bin/env', 'systemctl', 'disable', 'lightdm'],  # noqa: S603
                    check=True,
                )
    except Exception:
        logger.exception('Failed to handle SSH command.')
    return None



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os
from pathlib import Path

import dotenv

from ubo_app.error_handlers import setup_error_handling
from ubo_app.setup import setup

dotenv.load_dotenv(Path(__file__).parent / '.dev.env')


def setup_logging() -> None:
    from ubo_app.constants import GUI_LOG_LEVEL, LOG_LEVEL

    if LOG_LEVEL:
        import logging

        import ubo_app.logging

        level = getattr(
            ubo_app.logging,
            LOG_LEVEL,
            getattr(logging, LOG_LEVEL, logging.INFO),
        )

        ubo_app.logging.logger.setLevel(level)
        ubo_app.logging.add_file_handler(ubo_app.logging.logger, level)
        ubo_app.logging.add_stdout_handler(ubo_app.logging.logger, level)
    if GUI_LOG_LEVEL:
        import logging

        import ubo_gui.logger

        level = getattr(
            ubo_gui.logger,
            GUI_LOG_LEVEL,
            getattr(logging, GUI_LOG_LEVEL, logging.INFO),
        )

        ubo_gui.logger.logger.setLevel(level)
        ubo_gui.logger.add_file_handler(level)
        ubo_gui.logger.add_stdout_handler(level)


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    os.environ['KIVY_NO_CONFIG'] = '1'
    os.environ['KIVY_NO_FILELOG'] = '1'
    os.environ['KIVY_NO_CONSOLELOG'] = '1'
    os.environ['KCFG_KIVY_EXIT_ON_ESCAPE'] = '0'

    setup()
    setup_error_handling()
    setup_logging()

    from ubo_app.utils.loop import setup_event_loop

    setup_event_loop()

    import headless_kivy_pi.config

    headless_kivy_pi.config.setup_headless_kivy({'automatic_fps': True})

    from kivy.clock import Clock

    from ubo_app.load_services import load_services
    from ubo_app.menu import MenuApp

    load_services()
    app = MenuApp()

    try:
        app.run()
    finally:
        from redux import FinishAction

        from ubo_app.store import dispatch

        dispatch(FinishAction())

        # Needed since redux is scheduled using Clock scheduler and Clock doesn't run
        # after app is stopped.
        Clock.tick()


if __name__ == '__main__':
    main()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import sys
from dataclasses import field
from datetime import datetime, timezone
from enum import StrEnum, auto
from typing import Sequence
from uuid import uuid4

from immutable import Immutable
from redux import BaseAction, BaseEvent


class Importance(StrEnum):
    CRITICAL = auto()
    HIGH = auto()
    MEDIUM = auto()
    LOW = auto()


IMPORTANCE_ICONS = {
    Importance.CRITICAL: '󰅚',
    Importance.HIGH: '󰀪',
    Importance.MEDIUM: '',
    Importance.LOW: '󰌶',
}

IMPORTANCE_COLORS = {
    Importance.CRITICAL: '#D32F2F',
    Importance.HIGH: '#FFA000',
    Importance.MEDIUM: '#FFEB3B',
    Importance.LOW: '#2196F3',
}


class NotificationDisplayType(StrEnum):
    NOT_SET = auto()
    BACKGROUND = auto()
    FLASH = auto()
    STICKY = auto()


def default_icon() -> str:
    # WARNING: Dirty hack ahead
    # This is to set the default value of `icon` based on the provided/default value of
    # `importance`
    parent_frame = sys._getframe().f_back  # noqa: SLF001
    if not parent_frame:
        return ''
    return IMPORTANCE_ICONS[parent_frame.f_locals.get('importance', Importance.LOW)]


def default_color() -> str:
    # WARNING: Dirty hack ahead
    # This is to set the default value of `color` based on the provided/default value of
    # `importance`
    parent_frame = sys._getframe().f_back  # noqa: SLF001
    if not parent_frame:
        return ''
    return IMPORTANCE_COLORS[parent_frame.f_locals.get('importance', Importance.LOW)]


class Chime(StrEnum):
    ADD = 'add'
    DONE = 'done'
    FAILURE = 'failure'
    VOLUME_CHANGE = 'volume'


class Notification(Immutable):
    id: str = field(default_factory=lambda: uuid4().hex)
    title: str
    content: str
    importance: Importance = Importance.LOW
    chime: Chime = Chime.DONE
    timestamp: datetime = field(default_factory=lambda: datetime.now(tz=timezone.utc))
    is_read: bool = False
    sender: str | None = None
    actions: list[BaseAction | BaseEvent] = field(default_factory=list)
    icon: str = field(default_factory=default_icon)
    color: str = field(default_factory=default_color)
    expiry_date: datetime | None = None
    display_type: NotificationDisplayType = NotificationDisplayType.NOT_SET
    flash_time: float = 4


class NotificationsAction(BaseAction): ...


class NotificationsAddAction(NotificationsAction):
    notification: Notification


class NotificationsClearAction(NotificationsAction):
    notification: Notification


class NotificationsClearAllAction(NotificationsAction): ...


class NotificationsEvent(BaseEvent): ...


class NotificationsClearEvent(NotificationsEvent):
    notification: Notification


class NotificationsDisplayEvent(NotificationsEvent):
    notification: Notification


class NotificationsState(Immutable):
    notifications: Sequence[Notification]
    unread_count: int



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from pathlib import Path

from audio_manager import AudioManager
from constants import SOUND_MIC_STATE_ICON_ID, SOUND_MIC_STATE_ICON_PRIORITY

from ubo_app.store import autorun, dispatch, subscribe_event
from ubo_app.store.services.sound import SoundPlayChimeEvent
from ubo_app.store.status_icons import StatusIconsRegisterAction


def init_service() -> None:
    dispatch(
        StatusIconsRegisterAction(
            icon='󰍭',
            priority=SOUND_MIC_STATE_ICON_PRIORITY,
            id=SOUND_MIC_STATE_ICON_ID,
        ),
    )

    audio_manager = AudioManager()

    @autorun(lambda state: state.sound.playback_volume)
    def _(volume: float) -> None:
        audio_manager.set_playback_volume(volume)

    @autorun(lambda state: state.sound.capture_volume)
    def _(volume: float) -> None:
        audio_manager.set_capture_volume(volume)

    @autorun(lambda state: state.sound.is_playback_mute)
    def _(is_mute: bool) -> None:  # noqa: FBT001
        audio_manager.set_playback_mute(mute=is_mute)

    subscribe_event(
        SoundPlayChimeEvent,
        lambda event: audio_manager.play(
            Path(__file__).parent.joinpath(f'sounds/{event.name}.wav').as_posix(),
        ),
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import InitAction, InitializationActionError

from ubo_app.store.services.ssh import (
    SSHAction,
    SSHClearEnabledStateAction,
    SSHState,
    SSHUpdateStateAction,
)


def reducer(
    state: SSHState | None,
    action: SSHAction | InitAction,
) -> SSHState:
    if state is None:
        if isinstance(action, InitAction):
            return SSHState(is_active=False, is_enabled=False)
        raise InitializationActionError(action)

    if isinstance(action, SSHClearEnabledStateAction):
        return replace(state, is_enabled=None)

    if isinstance(action, SSHUpdateStateAction):
        if action.is_active is not None:
            state = replace(state, is_active=action.is_active)
        if action.is_enabled is not None:
            state = replace(state, is_enabled=action.is_enabled)
        return state
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.ip import (
    IpState,
    IpUpdateAction,
    IpUpdateRequestAction,
    IpUpdateRequestEvent,
)

Action = InitAction | IpUpdateAction | IpUpdateRequestAction
ResultEvent = IpUpdateRequestEvent


def reducer(
    state: IpState | None,
    action: Action,
) -> ReducerResult[IpState, Action, ResultEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return IpState(interfaces=[])
        raise InitializationActionError(action)

    if isinstance(action, IpUpdateAction):
        return replace(state, interfaces=action.interfaces)

    if isinstance(action, IpUpdateRequestAction):
        return CompleteReducerResult(
            state=replace(state, status=None) if action.reset else state,
            events=[IpUpdateRequestEvent()],
        )
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import sys
from dataclasses import replace
from pathlib import Path
from threading import current_thread, main_thread
from typing import TYPE_CHECKING, Callable, Coroutine

from redux import (
    BaseCombineReducerState,
    CombineReducerAction,
    CreateStoreOptions,
    InitAction,
    Store,
    combine_reducers,
)

from ubo_app.constants import DEBUG_MODE
from ubo_app.logging import logger
from ubo_app.store.main import MainAction, MainState
from ubo_app.store.main.reducer import reducer as main_reducer
from ubo_app.store.services.camera import CameraAction, CameraEvent
from ubo_app.store.services.docker import DockerAction, DockerState
from ubo_app.store.services.ip import IpAction, IpEvent, IpState
from ubo_app.store.services.keypad import KeypadEvent
from ubo_app.store.services.lightdm import LightDMAction, LightDMState
from ubo_app.store.services.notifications import NotificationsAction, NotificationsState
from ubo_app.store.services.rgb_ring import RgbRingAction
from ubo_app.store.services.sensors import SensorsAction, SensorsState
from ubo_app.store.services.sound import SoundAction, SoundState
from ubo_app.store.services.ssh import SSHAction, SSHState
from ubo_app.store.services.wifi import WiFiAction, WiFiEvent, WiFiState
from ubo_app.store.status_icons import StatusIconsAction, StatusIconsState
from ubo_app.store.status_icons.reducer import reducer as status_icons_reducer
from ubo_app.store.update_manager import UpdateManagerAction, UpdateManagerState
from ubo_app.store.update_manager.reducer import reducer as update_manager_reducer

if TYPE_CHECKING:
    from redux.basic_types import SnapshotAtom, TaskCreatorCallback

assert current_thread() is main_thread(), 'Store should be created in the main thread'  # noqa: S101


def scheduler(callback: Callable[[], None], *, interval: bool) -> None:
    from kivy.clock import Clock

    Clock.create_trigger(lambda _: callback(), 0, interval=interval)()


class RootState(BaseCombineReducerState):
    main: MainState
    lightdm: LightDMState
    status_icons: StatusIconsState
    update_manager: UpdateManagerState
    sensors: SensorsState
    ssh: SSHState
    sound: SoundState
    wifi: WiFiState
    ip: IpState
    notifications: NotificationsState
    docker: DockerState


ActionType = (
    CombineReducerAction
    | StatusIconsAction
    | UpdateManagerAction
    | MainAction
    | LightDMAction
    | SensorsAction
    | SSHAction
    | SoundAction
    | CameraAction
    | WiFiAction
    | IpAction
    | NotificationsAction
    | DockerAction
    | RgbRingAction
)
EventType = KeypadEvent | CameraEvent | WiFiEvent | IpEvent

root_reducer, root_reducer_id = combine_reducers(
    state_type=RootState,
    action_type=ActionType,  # pyright: ignore [reportArgumentType]
    event_type=EventType,  # pyright: ignore [reportArgumentType]
    main=main_reducer,
    status_icons=status_icons_reducer,
    update_manager=update_manager_reducer,
)


class UboStore(Store[RootState, ActionType, EventType]):
    @classmethod
    def serialize_value(cls: type[UboStore], obj: object | type) -> SnapshotAtom:
        from ubo_gui.menu.types import ActionItem
        from ubo_gui.page import PageWidget

        if isinstance(obj, type) and issubclass(obj, PageWidget):
            file_path = sys.modules[obj.__module__].__file__
            if file_path:
                return f"""{Path(file_path).relative_to(Path().absolute()).as_posix()}:{
                obj.__name__}"""
            return f'{obj.__module__}:{obj.__name__}'
        if isinstance(obj, ActionItem):
            obj = replace(obj, action='')
        return super().serialize_value(obj)


def create_task(
    coro: Coroutine,
    *,
    callback: TaskCreatorCallback | None = None,
) -> None:
    from ubo_app.utils.async_ import create_task

    create_task(coro, callback)


store = UboStore(
    root_reducer,
    CreateStoreOptions(
        auto_init=False,
        scheduler=scheduler,
        action_middlewares=[
            lambda action: logger.debug(
                'Action dispatched',
                extra={'action': action},
            )
            or action,
        ],
        event_middlewares=[
            lambda event: logger.debug(
                'Event dispatched',
                extra={'event': event},
            )
            or event,
        ],
        task_creator=create_task,
    ),
)

autorun = store.autorun
dispatch = store.dispatch
subscribe = store.subscribe
subscribe_event = store.subscribe_event

dispatch(InitAction())

if DEBUG_MODE:
    subscribe(lambda state: logger.verbose('State updated', extra={'state': state}))

__all__ = ('autorun', 'dispatch', 'subscribe', 'subscribe_event')



"""Module for sending commands to the system manager socket."""

from __future__ import annotations

import asyncio
import threading
from typing import Literal, overload

from ubo_app.constants import SERVER_SOCKET_PATH
from ubo_app.logging import logger

thread_lock = threading.Lock()


@overload
async def send_command(command: str) -> None: ...


@overload
async def send_command(command: str, *, has_output: Literal[True]) -> str: ...


async def send_command(command: str, *, has_output: bool = False) -> str | None:
    """Send a command to the system manager socket."""
    reader, writer = await asyncio.open_unix_connection(SERVER_SOCKET_PATH)

    logger.debug('Sending command:', extra={'command': command})

    response = None
    with thread_lock:
        writer.write(f'{command}'.encode() + b'\0')
        while has_output:
            datagram = (await reader.readuntil(b'\0'))[:-1]
            if not datagram:
                break
            response = datagram.decode('utf-8')
            logger.debug('Server response:', extra={'response': response})
        writer.close()
        await writer.wait_closed()

    logger.debug('Received response:', extra={'response': response})

    return response



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from kivy.utils import get_color_from_hex
from redux import (
    BaseEvent,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.notifications import (
    Importance,
    NotificationDisplayType,
    NotificationsAction,
    NotificationsAddAction,
    NotificationsClearAction,
    NotificationsClearAllAction,
    NotificationsClearEvent,
    NotificationsDisplayEvent,
    NotificationsState,
)
from ubo_app.store.services.rgb_ring import RgbRingBlinkAction
from ubo_app.store.services.sound import SoundPlayChimeAction

Action = InitAction | NotificationsAction
ResultAction = RgbRingBlinkAction | SoundPlayChimeAction


def reducer(
    state: NotificationsState | None,
    action: Action,
) -> ReducerResult[NotificationsState, ResultAction, BaseEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return NotificationsState(
                notifications=[],
                unread_count=0,
            )
        raise InitializationActionError(action)

    if isinstance(action, NotificationsAddAction):
        events = []
        if action.notification.display_type in (
            NotificationDisplayType.FLASH,
            NotificationDisplayType.STICKY,
        ):
            events.append(NotificationsDisplayEvent(notification=action.notification))
        if action.notification in state.notifications:
            return CompleteReducerResult(state=state, events=events)
        kivy_color = get_color_from_hex(action.notification.color)
        return CompleteReducerResult(
            state=replace(
                state,
                notifications=[
                    *[
                        notification
                        for notification in state.notifications
                        if notification.id != action.notification.id
                    ],
                    action.notification,
                ],
                unread_count=state.unread_count + 1,
            ),
            actions=[
                RgbRingBlinkAction(
                    color=(
                        round(kivy_color[0] * 256),
                        round(kivy_color[1] * 256),
                        round(kivy_color[2] * 256),
                    ),
                    repetitions={
                        Importance.LOW: 1,
                        Importance.MEDIUM: 2,
                        Importance.HIGH: 3,
                        Importance.CRITICAL: 4,
                    }[action.notification.importance],
                    wait=400,
                ),
                SoundPlayChimeAction(name=action.notification.chime),
            ],
            events=events,
        )
    if isinstance(action, NotificationsClearAction):
        return CompleteReducerResult(
            state=replace(
                state,
                notifications=[
                    notification
                    for notification in state.notifications
                    if notification is not action.notification
                ],
                unread_count=state.unread_count - 1
                if action.notification in state.notifications
                else state.unread_count,
            ),
            events=[NotificationsClearEvent(notification=action.notification)],
        )
    if isinstance(action, NotificationsClearAllAction):
        return replace(state, notifications=[], unread_count=0)
    return state



# ruff: noqa: BLE001, S112
"""Garbage collection investigation tools."""

from __future__ import annotations

import contextlib
import gc
from sys import stdout
from typing import TYPE_CHECKING, Callable

from redux.main import inspect

if TYPE_CHECKING:
    import weakref

SHORT_PRINT_LENGTH = 60


def short_print(obj: object) -> None:
    """Print the object."""
    stdout.write(str(type(obj)))
    try:
        stdout.write(
            str(obj)[:SHORT_PRINT_LENGTH] + '...'
            if len(str(obj)) > SHORT_PRINT_LENGTH
            else str(obj),
        )
        stdout.write('\n')
    except Exception as exception:
        if isinstance(exception, KeyboardInterrupt):
            raise
        stdout.write('Failed to print object\n')
    stdout.flush()


def examine(
    obj: object,
    *,
    depth_limit: int,
    filter_: Callable[[object], bool] = lambda _: True,
    looking_for: str | None = None,
) -> None:
    """Examine the object."""
    island = [obj]
    to_check: list[tuple[object, int]] = [(obj, 1)]
    path = []

    while to_check:
        obj, depth = to_check.pop()
        if looking_for:
            while len(path) >= depth:
                path.pop()
            path.append(obj)
        with contextlib.suppress(Exception):
            if looking_for and looking_for in str(type(obj)):
                stdout.write('Found\n')
                stdout.flush()
                for i in path:
                    short_print(i)
                break
        try:
            referrers = [
                i
                for i in gc.get_referrers(obj)
                if i not in island
                and i not in to_check
                and i is not path
                and i is not island
                and i is not to_check
            ]
        except Exception:
            continue

        for i in referrers:
            island.append(i)
            if filter_(i):
                short_print(i)
            if depth < depth_limit:
                to_check.append((i, depth + 1))
        del referrers


def search_stack_for_instance(ref: weakref.ReferenceType) -> None:
    """Search the stack for an instance."""
    frames = inspect.stack()
    with contextlib.suppress(Exception):
        for frame in frames:
            local_vars = frame.frame.f_locals
            for var_name, var_value in local_vars.items():
                if var_value is ref():
                    stdout.write(
                        'Found instance'
                        + str(
                            {
                                'var_name': var_name,
                                'var_value': var_value,
                                'frame': frame,
                            },
                        )
                        + '\n',
                    )
                    stdout.flush()



"""Setup the service."""

from __future__ import annotations

import asyncio
import contextlib
from dataclasses import fields
from pathlib import Path
from typing import Callable

import docker
import docker.errors
from docker.models.containers import Container
from docker.models.images import Image
from ubo_gui.menu.types import ActionItem, HeadedMenu, HeadlessMenu, Item, SubMenuItem

from ubo_app.constants import DOCKER_INSTALLATION_LOCK_FILE, SERVER_SOCKET_PATH
from ubo_app.store import autorun, dispatch
from ubo_app.store.main import RegisterRegularAppAction
from ubo_app.store.services.docker import (
    DockerSetStatusAction,
    DockerState,
    DockerStatus,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.monitor_unit import monitor_unit
from ubo_app.utils.server import send_command


def install_docker() -> None:
    """Install Docker."""
    if Path(SERVER_SOCKET_PATH).exists():

        async def act() -> None:
            await send_command('docker install')
            dispatch(DockerSetStatusAction(status=DockerStatus.INSTALLING))

        create_task(act())


def run_docker() -> None:
    """Install Docker."""

    async def act() -> None:
        await send_command('docker start')
        dispatch(DockerSetStatusAction(status=DockerStatus.UNKNOWN))

    create_task(act())


def stop_docker() -> None:
    """Install Docker."""

    async def act() -> None:
        await send_command('docker stop')
        dispatch(DockerSetStatusAction(status=DockerStatus.UNKNOWN))

    create_task(act())


async def check_docker() -> None:
    """Check if Docker is installed."""
    from image import update_container
    from reducer import IMAGES

    process = await asyncio.create_subprocess_exec(
        '/usr/bin/env',
        'which',
        'docker',
    )
    await process.wait()
    is_installed = process.returncode == 0

    is_running = False
    with contextlib.suppress(Exception):
        docker_client = docker.from_env()
        is_running = docker_client.ping()

        for container in docker_client.containers.list(all=True):
            if not isinstance(container, Container):
                continue

            with contextlib.suppress(docker.errors.DockerException):
                container_image = container.image
                for image_id, image_description in IMAGES.items():
                    if (
                        isinstance(container_image, Image)
                        and image_description.path in container_image.tags
                    ):
                        update_container(image_id, container)

        docker_client.close()

    if is_running:
        dispatch(DockerSetStatusAction(status=DockerStatus.RUNNING))
    elif is_installed:
        dispatch(DockerSetStatusAction(status=DockerStatus.NOT_RUNNING))
    elif Path(DOCKER_INSTALLATION_LOCK_FILE).exists():
        dispatch(DockerSetStatusAction(status=DockerStatus.INSTALLING))
    else:
        dispatch(DockerSetStatusAction(status=DockerStatus.NOT_INSTALLED))


@autorun(lambda state: state.docker.service.status)
def setup_menu(status: DockerStatus) -> HeadedMenu:
    """Get the menu items for the Docker service."""
    title = 'Setup Docker'
    if status == DockerStatus.UNKNOWN:
        return HeadedMenu(
            title=title,
            heading='Checking',
            sub_heading='Checking Docker service status',
            items=[],
            placeholder='',
        )
    if status == DockerStatus.NOT_INSTALLED:
        return HeadedMenu(
            title=title,
            heading='Docker is not Installed',
            sub_heading='Install it to enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Install Docker',
                    icon='󰶮',
                    action=install_docker,
                ),
            ],
        )
    if status == DockerStatus.INSTALLING:
        return HeadedMenu(
            title=title,
            heading='Installing...',
            sub_heading='Docker is being installed',
            items=[],
            placeholder='',
        )
    if status == DockerStatus.NOT_RUNNING:
        return HeadedMenu(
            title=title,
            heading='Docker is not Running',
            sub_heading='Run it to enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Start Docker',
                    icon='󰐊',
                    action=run_docker,
                ),
            ],
        )
    if status == DockerStatus.RUNNING:
        return HeadedMenu(
            title=title,
            heading='Docker is Running',
            sub_heading='Enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Stop Docker',
                    icon='󰓛',
                    action=stop_docker,
                ),
            ],
        )
    if status == DockerStatus.ERROR:
        return HeadedMenu(
            title=title,
            heading='Docker Error',
            sub_heading='Please check the logs for more information',
            items=[],
            placeholder='',
        )

    msg = f'Unknown status: {status}'
    raise ValueError(msg)


def setup_menu_action() -> Callable[[], HeadedMenu]:
    """Get the menu items for the Docker service."""
    create_task(check_docker())
    return setup_menu


@autorun(lambda state: state.docker)
def docker_menu_items(state: DockerState) -> list[Item]:
    """Get the menu items for the Docker service."""
    create_task(check_docker())
    items: list[Item] = [
        ActionItem(
            label='Setup Docker',
            icon='',
            action=setup_menu_action,
        ),
    ]

    if state.service.status == DockerStatus.RUNNING:
        from image import image_menus

        items.append(
            SubMenuItem(
                label='Docker Containers',
                icon='󱣘',
                sub_menu=HeadlessMenu(
                    title='Docker Containers',
                    items=[
                        ActionItem(
                            label=getattr(state, image_id).label,
                            icon=getattr(state, image_id).icon,
                            action=image_menus[image_id],
                        )
                        for image_id in (field.name for field in fields(state))
                        if image_id not in (field.name for field in fields(DockerState))
                    ],
                    placeholder='No Docker containers',
                ),
            ),
        )

    return items


def docker_menu_item_action() -> HeadlessMenu:
    """Get the menu items for the Docker service."""
    return HeadlessMenu(
        title='Docker',
        items=docker_menu_items,
    )


docker_main_menu = ActionItem(
    label='Docker',
    icon='󰡨',
    action=docker_menu_item_action,
)


def init_service() -> None:
    """Initialize the service."""
    dispatch(RegisterRegularAppAction(menu_item=docker_main_menu))
    create_task(
        monitor_unit(
            'docker.socket',
            lambda status: dispatch(
                DockerSetStatusAction(
                    status=DockerStatus.RUNNING
                    if status in ('active', 'activating', 'reloading')
                    else DockerStatus.NOT_RUNNING,
                ),
            ),
        ),
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction


class SSHAction(BaseAction): ...


class SSHUpdateStateAction(SSHAction):
    is_active: bool | None = None
    is_enabled: bool | None = None


class SSHClearEnabledStateAction(SSHAction): ...


class SSHState(Immutable):
    is_active: bool = False
    is_enabled: bool | None = None



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import contextlib
from dataclasses import replace

import semver
from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)
from ubo_gui.constants import SECONDARY_COLOR

from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.update_manager import (
    UPDATE_MANAGER_NOTIFICATION_ID,
    UpdateManagerAction,
    UpdateManagerCheckEvent,
    UpdateManagerEvent,
    UpdateManagerSetStatusAction,
    UpdateManagerSetVersionsAction,
    UpdateManagerState,
    UpdateManagerUpdateEvent,
    UpdateStatus,
)

ABOUT_MENU_PATH = ['Dashboard', 'Main', 'About']


def reducer(
    state: UpdateManagerState | None,
    action: UpdateManagerAction,
) -> ReducerResult[
    UpdateManagerState,
    UpdateManagerSetStatusAction | NotificationsAddAction,
    UpdateManagerEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            return UpdateManagerState()
        raise InitializationActionError(action)

    if isinstance(action, UpdateManagerSetVersionsAction):
        state = replace(
            state,
            current_version=action.current_version,
            latest_version=action.latest_version,
        )
        version_comparison = 1
        with contextlib.suppress(ValueError):
            latest_version = semver.Version.parse(
                action.latest_version,
                optional_minor_and_patch=True,
            )
            current_version = semver.Version.parse(
                action.current_version,
                optional_minor_and_patch=True,
            )
            version_comparison = latest_version.compare(current_version)
        if version_comparison > 0:
            return CompleteReducerResult(
                state=state,
                actions=[
                    UpdateManagerSetStatusAction(status=UpdateStatus.OUTDATED),
                    NotificationsAddAction(
                        notification=Notification(
                            id=UPDATE_MANAGER_NOTIFICATION_ID,
                            title='Update available!',
                            content=f"""Ubo v{action.latest_version
                            } is available. Go to the About menu to update.""",
                            display_type=NotificationDisplayType.FLASH
                            if action.flash_notification
                            else NotificationDisplayType.BACKGROUND,
                            color=SECONDARY_COLOR,
                            icon='󰬬',
                            chime=Chime.DONE,
                        ),
                    ),
                ],
            )
        return CompleteReducerResult(
            state=state,
            actions=[UpdateManagerSetStatusAction(status=UpdateStatus.UP_TO_DATE)],
        )

    if isinstance(action, UpdateManagerSetStatusAction):
        events = []
        if action.status == UpdateStatus.CHECKING:
            events.append(UpdateManagerCheckEvent())
        elif action.status == UpdateStatus.UPDATING:
            events.append(UpdateManagerUpdateEvent())

        return CompleteReducerResult(
            state=replace(
                state,
                update_status=action.status,
            ),
            events=events,
        )

    return state



"""LightDM service module."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Sequence

from ubo_gui.menu.types import ActionItem, HeadlessMenu, Item, Menu

from ubo_app.store import autorun, dispatch
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.lightdm import (
    LightDMClearEnabledStateAction,
    LightDMUpdateStateAction,
)
from ubo_app.utils.async_ import create_task
from ubo_app.utils.monitor_unit import is_unit_active, is_unit_enabled, monitor_unit
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from ubo_app.store.services.lightdm import LightDMState


def start_lightdm_service() -> None:
    """Start the LightDM service."""

    async def act() -> None:
        await send_command('service lightdm start')

    create_task(act())


def stop_lightdm_service() -> None:
    """Stop the LightDM service."""

    async def act() -> None:
        await send_command('service lightdm stop')

    create_task(act())


def enable_lightdm_service() -> None:
    """Enable the LightDM service."""

    async def act() -> None:
        dispatch(LightDMClearEnabledStateAction())
        await send_command('service lightdm enable')
        await asyncio.sleep(5)
        await check_is_lightdm_enabled()

    create_task(act())


def disable_lightdm_service() -> None:
    """Disable the LightDM service."""

    async def act() -> None:
        dispatch(LightDMClearEnabledStateAction())
        await send_command('service lightdm disable')
        await asyncio.sleep(5)
        await check_is_lightdm_enabled()

    create_task(act())


@autorun(lambda state: state.lightdm)
def lightdm_items(state: LightDMState) -> Sequence[Item]:
    """Get the LightDM menu items."""
    return [
        ActionItem(
            label='Stop' if state.is_active else 'Start',
            icon='󰓛' if state.is_active else '󰐊',
            action=stop_lightdm_service if state.is_active else start_lightdm_service,
        ),
        Item(
            label='...',
            icon='',
        )
        if state.is_enabled is None
        else ActionItem(
            label='Disable',
            icon='[color=#008000]󰯄[/color]',
            action=disable_lightdm_service,
        )
        if state.is_enabled
        else ActionItem(
            label='Enable',
            icon='[color=#ffff00]󰯅[/color]',
            action=enable_lightdm_service,
        ),
    ]


@autorun(lambda state: state.lightdm)
def lightdm_icon(state: LightDMState) -> str:
    """Get the LightDM icon."""
    return '[color=#008000]󰪥[/color]' if state.is_active else '[color=#ffff00]󰝦[/color]'


@autorun(lambda state: state.lightdm)
def lightdm_title(_: LightDMState) -> str:
    """Get the LightDM title."""
    return lightdm_icon() + ' LightDM'


async def check_is_lightdm_active() -> None:
    """Check if the LightDM service is active."""
    if await is_unit_active('lightdm'):
        dispatch(LightDMUpdateStateAction(is_enabled=True))
    else:
        dispatch(LightDMUpdateStateAction(is_enabled=False))


async def check_is_lightdm_enabled() -> None:
    """Check if the LightDM service is enabled."""
    if await is_unit_enabled('lightdm'):
        dispatch(LightDMUpdateStateAction(is_enabled=True))
    else:
        dispatch(LightDMUpdateStateAction(is_enabled=False))


def open_lightdm_menu() -> Menu:
    """Open the LightDM menu."""
    create_task(
        asyncio.gather(
            check_is_lightdm_active(),
            check_is_lightdm_enabled(),
        ),
    )

    return HeadlessMenu(
        title=lightdm_title,
        items=lightdm_items,
    )


def init_service() -> None:
    """Initialize the LightDM service."""
    dispatch(
        RegisterSettingAppAction(
            menu_item=ActionItem(
                label='LightDM',
                icon=lightdm_icon,
                action=open_lightdm_menu,
            ),
        ),
    )

    create_task(
        asyncio.gather(
            check_is_lightdm_active(),
            check_is_lightdm_enabled(),
            monitor_unit(
                'lightdm.service',
                lambda status: dispatch(
                    LightDMUpdateStateAction(
                        is_active=status in ('active', 'activating', 'reloading'),
                    ),
                ),
            ),
        ),
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107, PLR2004
from __future__ import annotations

import grp
import logging
import os
import pwd
import socket
import stat
import sys
from pathlib import Path
from threading import Thread

from ubo_app.constants import USERNAME
from ubo_app.error_handlers import setup_error_handling
from ubo_app.logging import add_file_handler, add_stdout_handler, get_logger
from ubo_app.system.system_manager.docker import docker_handler
from ubo_app.system.system_manager.led import LEDManager
from ubo_app.system.system_manager.service_manager import system_handler

SOCKET_PATH = Path(os.environ.get('RUNTIME_DIRECTORY', '/run/ubo')).joinpath(
    'system_manager.sock',
)

led_manager = LEDManager()

logger = get_logger('system-manager')
add_file_handler(logger, logging.DEBUG)
add_stdout_handler(logger, logging.DEBUG)
logger.setLevel(logging.DEBUG)


def handle_command(command: str) -> str | None:
    header, *incoming = command.split()
    if header == 'led':
        led_manager.run_command_thread_safe(incoming)
    elif header == 'docker':
        thread = Thread(target=docker_handler, args=(incoming[0],))
        thread.start()
    elif header == 'service':
        return system_handler(incoming[0], incoming[1])
    return None


def main() -> None:
    """Initialise the System-Manager."""
    setup_error_handling()
    logger.debug('Initialising System-Manager...')

    led_manager.run_command_thread_safe('spinning_wheel 255 255 255 50 6 100'.split())

    uid = pwd.getpwnam('root').pw_uid
    gid = grp.getgrnam(USERNAME).gr_gid

    SOCKET_PATH.unlink(missing_ok=True)
    SOCKET_PATH.parent.mkdir(parents=True, exist_ok=True)

    logger.info('System Manager opening socket...')
    server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    server.bind(SOCKET_PATH.as_posix())

    server.listen()
    permission = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWGRP | stat.S_IWUSR
    )
    SOCKET_PATH.chmod(permission)

    os.chown(SOCKET_PATH, uid, gid)

    logger.info('System Manager Listening...')
    remaining = b''
    while True:
        try:
            connection, client_address = server.accept()
            logger.debug('New connection:', extra={'client_address': client_address})
            datagram = remaining + connection.recv(1024)
            if not datagram:
                break
            if b'\0' not in datagram:
                remaining = datagram
                continue

            command, remaining = datagram.split(b'\0', 1)

            logger.debug('Received command:', extra={'command': command})
            result = handle_command(command.decode('utf-8'))
            if result is not None:
                connection.sendall(result.encode() + b'\0')
                connection.close()

        except KeyboardInterrupt:
            logger.debug('Interrupted')
            server.close()
            try:
                sys.exit(0)
            except SystemExit:
                os._exit(0)
    logger.info('Shutting down...')
    server.close()
    SOCKET_PATH.unlink()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import InitAction, InitializationActionError

from ubo_app.store.services.lightdm import (
    LightDMAction,
    LightDMClearEnabledStateAction,
    LightDMState,
    LightDMUpdateStateAction,
)


def reducer(
    state: LightDMState | None,
    action: LightDMAction | InitAction,
) -> LightDMState:
    if state is None:
        if isinstance(action, InitAction):
            return LightDMState(is_active=False, is_enabled=False)
        raise InitializationActionError(action)

    if isinstance(action, LightDMClearEnabledStateAction):
        return replace(state, is_enabled=None)

    if isinstance(action, LightDMUpdateStateAction):
        if action.is_active is not None:
            state = replace(state, is_active=action.is_active)
        if action.is_enabled is not None:
            state = replace(state, is_enabled=action.is_enabled)
        return state
    return state



"""Module for scanning QR codes using the camera."""

from __future__ import annotations

import asyncio
import uuid
from asyncio import Future
from datetime import datetime, timezone
from typing import Callable, TypeAlias, overload

from typing_extensions import TypeVar

from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.services.camera import (
    CameraBarcodeEvent,
    CameraStartViewfinderAction,
    CameraStopViewfinderEvent,
)
from ubo_app.store.services.notifications import (
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
    NotificationsClearEvent,
)
from ubo_app.store.services.rgb_ring import RgbRingBlinkAction

QrCodeGroupDict: TypeAlias = dict[str, str | None] | None


ReturnType = TypeVar('ReturnType', infer_variance=True)


@overload
async def qrcode_input(
    pattern: str,
    *,
    prompt: str | None = None,
) -> tuple[str, QrCodeGroupDict]: ...


@overload
async def qrcode_input(
    pattern: str,
    *,
    prompt: str | None = None,
    resolver: Callable[[str, QrCodeGroupDict], ReturnType],
) -> ReturnType: ...


async def qrcode_input(
    pattern: str,
    *,
    prompt: str | None = None,
    resolver: Callable[[str, QrCodeGroupDict], ReturnType] | None = None,
) -> tuple[str, QrCodeGroupDict] | ReturnType:
    """Use the camera to scan a QR code."""
    future = Future[tuple[str, QrCodeGroupDict]]()
    id = uuid.uuid4().hex
    loop = asyncio.get_running_loop()

    if prompt:
        notification_future = Future()
        notification = Notification(
            id='qrcode',
            title='QR Code',
            content=prompt,
            display_type=NotificationDisplayType.STICKY,
            is_read=True,
            expiry_date=datetime.now(tz=timezone.utc),
        )

        def clear_notification(event: NotificationsClearEvent) -> None:
            if event.notification == notification:
                loop.call_soon_threadsafe(notification_future.set_result, None)

        subscribe_event(
            NotificationsClearEvent,
            clear_notification,
        )
        dispatch(NotificationsAddAction(notification=notification))

        await notification_future

    async def handle_barcode_event(event: CameraBarcodeEvent) -> None:
        if event.id == id:
            from kivy.utils import get_color_from_hex

            loop.call_soon_threadsafe(future.set_result, (event.code, event.group_dict))
            kivy_color = get_color_from_hex('#21E693')
            dispatch(
                RgbRingBlinkAction(
                    color=(
                        round(kivy_color[0] * 256),
                        round(kivy_color[1] * 256),
                        round(kivy_color[2] * 256),
                    ),
                    repetitions=1,
                    wait=200,
                ),
            )

    def handle_cancel(event: CameraStopViewfinderEvent) -> None:
        if event.id == id:
            future.cancel()

    subscribe_event(
        CameraBarcodeEvent,
        handle_barcode_event,
        keep_ref=False,
    )
    subscribe_event(
        CameraStopViewfinderEvent,
        handle_cancel,
    )
    dispatch(CameraStartViewfinderAction(id=id, pattern=pattern))

    result = await future

    if not resolver:
        return result

    return resolver(*result)
